/* This is under the same copyright as Brian Gladman's code, except that
   I, Sam Trenholme, own the copyright.  */

/* Includes */
#include "aes.h"
#include <stdlib.h>

/* Maximum allowed block and key size, in octets */

#define MAXBLOCKSIZE 32
#define MAXKEYSIZE 32

/* Convert an ascii hex string in to a binary string.
   Input: Pointer to hex string (NULL-terminated), pointer to binary 
          string, maximum allowed length of binary string in bytes
   Output: Length of binary string in bytes, -1 on error
*/

int ascii_to_bin(char *ascii, char *bin, int maxlen) {
    int counter;
    char left, right, nibble, byt;

    counter = 0;

    while(*ascii != NULL && counter < maxlen) {
        /* Get the left nibble of this byte */
	left = *ascii;
	/* Convert to lower case if needed */
	if(left >= 'A' && left <= 'Z')
	    left += 32;
        /* Make sure the left nibble is in bounds. */
	if(left < '0' || (left > '9' && left < 'a') || left > 'f')
	    return -1;
        
	/* Now, do the same for the right nibble */
	ascii++;
	right = *ascii;
	if(right >= 'A' && right <= 'Z')
	    right += 32;
        if(right < '0' || (right > '9' && right < 'a') || right > 'f')
	    return -1;
        ascii++;

	/* Now, make the nibbles a byte */
        if(left <= '9')
	    nibble = left - '0';
        else
	    nibble = left + 10 - 'a';
        nibble <<= 4;
	if(right <= '9')
	    byt = right - '0';
        else
	    byt = right + 10 - 'a';
        byt |= nibble;

        *(bin + counter) = byt;
	counter++;
        }

    return counter;
    }

/* Test the AES cipher with a given key and input block.
   Input: ASCII hex key and input
   Output: -1 on fail, 1 on success, show key, input, and ciphertext on
           standard output
 */

int test_aes(char *key, char *input) {
 
    char binkey[MAXKEYSIZE + 3], bininput[MAXBLOCKSIZE + 3];
    char binoutput[MAXBLOCKSIZE + 3];
    aes cx;
    int keylen, blocklen; /* Length of key and block, in bytes */
    int counter;
    
    /* Initialize the context of the AES cipher */
    cx.mode = 0;

    /* Convert the key and input in to binary data */
    keylen = ascii_to_bin(key,binkey,MAXKEYSIZE);
    blocklen = ascii_to_bin(input,bininput,MAXBLOCKSIZE);

    if(keylen == -1 || blocklen == -1)
        return -1;

    /* Show the end user the key and plaintext */
    printf("   key     = ");
    for(counter = 0; counter < keylen; counter++)
        printf("%02x",binkey[counter] & 0xff);
    printf("\n");
    printf("   input   = ");
    for(counter = 0; counter < blocklen; counter++)
        printf("%02x",bininput[counter] & 0xff);
    printf("\n");

    /* Set the length of the block */
    set_blk(blocklen,&cx);
    /* Set up the encryption key */
    set_key(binkey,keylen,enc,&cx);
    /* Encrypt the block of plaintext in to ciphertext */
    encrypt(bininput,binoutput,&cx);

    /* Show the end user the ciphertext */
    printf("   enc     = ");
    for(counter = 0; counter < blocklen; counter++)
        printf("%02x",binoutput[counter] & 0xff);
    printf("\n\n");

    return 1;

    }

main() {
    /* 128-bit key and block size */
    test_aes("2b7e151628aed2a6abf7158809cf4f3c",
             "3243f6a8885a308d313198a2e0370734");
    test_aes("00ffeeddccbbaa998877665544332211",
             "112233445566778899aabbccddeeff00");
    test_aes("ffeeddccbbaa99887766554433221100",
             "00112233445566778899aabbccddeeff");
    test_aes("00000000000000000000000000000000",
             "00000000000000000000000000000000");
    test_aes("ffffffffffffffffffffffffffffffff",
             "00000000000000000000000000000000");
    test_aes("ffffffffffffffffffffffffffffffff",
             "ffffffffffffffffffffffffffffffff");
    test_aes("000102030405060708090a0b0c0d0e0f",
             "000102030405060708090a0b0c0d0e0f");
    test_aes("31393730303132353038363735333039",
             "30303030303030303030303030303030");
    /* 160-bit key and block size */
    test_aes("0000000000000000000000000000000000000000",
             "0000000000000000000000000000000000000000");
    test_aes("2b7e151628aed2a6abf7158809cf4f3c762e7160",
             "3243f6a8885a308d313198a2e03707344a409382");
    }

    
