// The FastHash256 File Hashing Program

#include <stdio.h>
#include "Interface.h"
#include "Qbox.h"

int main(int argc, char **argv)
{
     // 192-bit Key and Internal State
     unsigned long int W[8], in_data[8];
     
     // Char Ptr to in_data
     char *in_char;
     
     // Stream Counter and Tmp Variable
     unsigned long int stream_count, tmp;
     
     unsigned short int i, stream_left;
     
     // Input File Name, Output File Name, Flag
     char *in_file, flag = 0;
     
     FILE *fp_in;
         
     if(argc != 2) {printf ("Usage: fastflexhash {input file}\n");
                    exit(0);
                   } 

     in_file = argv[1];

     // Open Input File for Reading
     fp_in = fopen(in_file, "rb");
     if(!fp_in)
     {
          printf("Could Not Open File: %s", in_file);
          getchar();
          exit(-1);
     }
     
     // Size of Input File
     fseek(fp_in,0,SEEK_END);
     stream_count = ftell(fp_in);
     /* printf("Hashing %d Bytes\n", stream_count); */
     
     rewind(fp_in);
     
     // Input File Size is a Multiple of 256 bits
     stream_left = stream_count % 32 ;
    
     if(!(stream_left))
     {
          flag = '1';
     }
     else if(stream_left <= 28)
          {
                flag = '2';
          }
          else
             {
                  flag = '3';
             }
    

     /* The specification for FastHash does not specify anywhere what values
      * we should initialize the hash state with.  I'm using the the hex 
      * values for the square roots of 2,3,5,7,11,13,17, and 19, all
      * minus one */
     W[0] = 0x6a09e667UL; /* Square root of 2 minus one */
     W[1] = 0x428a2f98UL;
     W[2] = 0x2611186bUL;
     W[3] = 0x1aa59c41UL;
     W[4] = 0x10a68868UL;
     W[5] = 0x0e05264bUL;
     W[6] = 0x0aa758ccUL;
     W[7] = 0x0982fe34UL;

     for(tmp = 0; tmp < stream_count/32; tmp++)
     {
                           
          fread(in_data,4,8,fp_in);
          
          for(i=0; i<8; i++)
          {
               W[i] ^= in_data[i]; 
          }
          
          FastFlex(W);              
     }
     
     
     switch(flag)
     {
           
           case '1':
                {
                     for(i=0; i<7; i++)
                     in_data[i] = 0;
                     
                     in_data[7] = stream_count;
                     
                     for(i=0; i<8; i++)
                     {
                          W[i] ^= in_data[i]; 
                     }
                     
                     FastFlex(W);
                     
                     break;
                }
                
           case '2':
                {
                     fread(in_data,1,stream_left,fp_in);
                     
                     in_char = (char *) in_data;
                     in_char += stream_left;
                     
                     for(i=stream_left; i<28; i++)
                     *in_char = 0;
                     
                     in_data[7] = stream_count;
                     
                     for(i=0; i<8; i++)
                     {
                          W[i] ^= in_data[i]; 
                     }
                     
                     FastFlex(W);
                      
                     break;
                }
                
           case '3':
                {
                     fread(in_data,1,stream_left,fp_in);
                     
                     in_char = (char *) in_data;
                     in_char += stream_left;
                     
                     for(i=stream_left; i<32; i++)
                     *in_char = 0;
                     
                     FastFlex(W);
                     
                     W[7] ^= stream_count;
                     
                     FastFlex(W);               
                     break;
                }           
     }
   
     unsigned char *z;
     z = (char *)W;
     for(i = 0; i < 32; i++) {
	printf("%02x",z[i]);
     } 
     printf(" %s\n",argv[1]);
     
     fclose(fp_in);
     /*getch();*/
}


