// Interface2.h for Intel

// The Timing Function using the RDTSC Instruction
inline unsigned long long int rdtsc(void)
{
     unsigned long long int x;
     asm volatile (".byte 0x0f, 0x31" : "=A" (x));
     return x;
}


// FastFlex, FastStream192, FastHash256, KeyHash256 & QuickStream192 Interfaces

#define r_fast 8            // 8 Rounds for Fast Functions
#define r_quick 4           // 4 Rounds for Quick Functions
#define constant_xor        0x8b9a7465

// Thge Union for 64-bit -- 2*32-bit Words used in Nonce / col_1 and col_2
union nonce_struct {
                             unsigned long long int counter;
                             unsigned long int word[2];
                        };

extern unsigned long int Qbox[];

// The size of the keystream buffer
#define stream_buffsize 256

//The Macro for Left Rotation
#define RL(x,y)       (((x) << (y)) | ((x) >> (32-(y))))

// The Macro for S() function
#define S(x)          ((x)^Qbox[(x)>>24])



// *****************************************************************************
// Interface for FastFlex()
inline void FastFlex(unsigned long int *);

//The FastFlex() Function
inline void FastFlex(unsigned long int W[8])
{
     unsigned long int tmp_san; 
     // One Temporary Variable Needed in the SAN 
     
     unsigned long long int col_1, col_2;
     // The col_1 and col_2 variables hold the 64-bit product Y(I)*Y(I+4)
     // temporarily. They are used to improve pipeline performance.
     
     unsigned long int i;
     // Round Counter for HashCore
     
     // The HashCore starts here
      for(i=0; i<r_fast; i++ )
     {
              // Calculate W[1] W[5],
              W[1] ^= RL(W[0] + W[3], 7);
              W[7] ^= RL(W[6] + W[5], 7);       
              
              // Calculate W[2] W[4],
              W[2] ^= RL((W[1] + W[0]), 13);
              W[4] ^= RL((W[7] + W[6]), 13); 
              
              // Calculate W[3] W[5], 
              W[3] ^= RL((W[2] + W[1]), 17);
              W[5] ^= RL((W[4] + W[7]), 17);
                      
              // Calculate W[0] W[6], 
              W[0] ^= RL((W[3] + W[2]), 23);
              W[6] ^= RL((W[5] + W[4]), 23);
              
              col_1 = (unsigned long long int)W[1] * (unsigned long long int)W[7];
              W[1] = col_1;
              W[7] = (col_1>>32)^constant_xor;
              
              col_2 = (unsigned long long int)W[2] * (unsigned long long int)W[4];
              W[2] = col_2;
              W[4] = (col_2>>32)^constant_xor; 
                           
              col_1 = (unsigned long long int)W[3] * (unsigned long long int)W[5];
              W[3] = col_1;
              W[5] = (col_1>>32)^constant_xor;         
              
              col_2 = (unsigned long long int)W[0] * (unsigned long long int)W[6];      
              W[0] = col_2;
              W[6] = (col_2>>32)^constant_xor;         
         
     }
     
     
     // The Substitution Addition Network.
     // A = W[0], B = W[1], C = W[2], D = W[3]
     // E = W[4], F = W[5], G = W[6], H = W[7]  
     //Layer1
              tmp_san = W[0];
              W[0] += S(W[1]) + W[5];
              W[1] += S(W[2]) + W[6];
              W[2] += S(W[3]) + W[7];
              W[3] += S(tmp_san) + W[4];
             
     //Layer2
              W[4] += S(W[0]);
              W[5] += S(W[1]);
              W[6] += S(W[2]);
              W[7] += S(W[3]);
             
     //Layer3
              W[0] ^= S(W[6]);
              W[1] ^= S(W[7]);
              W[2] ^= S(W[4]);
              W[3] ^= S(W[5]);
             
     //Layer5
              tmp_san = W[4];
              W[4] += W[0] ^= W[5];
              W[5] += W[1] ^= W[6];
              W[6] += W[2] ^= W[7];
              W[7] += W[3] ^= tmp_san; 
          
}


//******************************************************************************
// Interface for QuickFlex()
inline void QuickFlex(unsigned long int *);

//The QuickFlex() Function
inline void QuickFlex(unsigned long int W[8])
{
     unsigned long int tmp_san; 
     // One Temporary Variable Needed in the SAN 
     
     unsigned long long int col_1, col_2;
     // The col_1 and col_2 variables hold the 64-bit product Y(I)*Y(I+4)
     // temporarily. They are used to improve pipeline performance.
     
     unsigned long int i;
     // Round Counter for HashCore
     
     // The HashCore starts here
     for(i=0; i<r_quick; i++ )
     {
              // Calculate W[1] W[5],
              W[1] ^= RL(W[0] + W[3], 7);
              W[7] ^= RL(W[6] + W[5], 7);       
              
              // Calculate W[2] W[4],
              W[2] ^= RL((W[1] + W[0]), 13);
              W[4] ^= RL((W[7] + W[6]), 13); 
              
              // Calculate W[3] W[5], 
              W[3] ^= RL((W[2] + W[1]), 17);
              W[5] ^= RL((W[4] + W[7]), 17);
                      
              // Calculate W[0] W[6], 
              W[0] ^= RL((W[3] + W[2]), 23);
              W[6] ^= RL((W[5] + W[4]), 23);
              
              col_1 = (unsigned long long int)W[1] * (unsigned long long int)W[7];
              W[1] = col_1;
              W[7] = (col_1>>32)^constant_xor;
              
              col_2 = (unsigned long long int)W[2] * (unsigned long long int)W[4];
              W[2] = col_2;
              W[4] = (col_2>>32)^constant_xor; 
                           
              col_1 = (unsigned long long int)W[3] * (unsigned long long int)W[5];
              W[3] = col_1;
              W[5] = (col_1>>32)^constant_xor;         
              
              col_2 = (unsigned long long int)W[0] * (unsigned long long int)W[6];      
              W[0] = col_2;
              W[6] = (col_2>>32)^constant_xor;         
         
     }
         
     // The Substitution Addition Network.
     // A = W[0], B = W[1], C = W[2], D = W[3]
     // E = W[4], F = W[5], G = W[6], H = W[7]  
     //Layer1
              tmp_san = W[0];
              W[0] += S(W[1]) + W[5];
              W[1] += S(W[2]) + W[6];
              W[2] += S(W[3]) + W[7];
              W[3] += S(tmp_san) + W[4];
             
     //Layer2
              W[4] += S(W[0]);
              W[5] += S(W[1]);
              W[6] += S(W[2]);
              W[7] += S(W[3]);
             
     //Layer3
              W[0] ^= S(W[6]);
              W[1] ^= S(W[7]);
              W[2] ^= S(W[4]);
              W[3] ^= S(W[5]);
             
     //Layer5
              tmp_san = W[4];
              W[4] += W[0] ^= W[5];
              W[5] += W[1] ^= W[6];
              W[6] += W[2] ^= W[7];
              W[7] += W[3] ^= tmp_san;        
          
}   

// *****************************************************************************


