#include "benchmark.h"

inline uint64 grab_tod()
{
  struct timeval  tv;
  struct timezone tz;

  gettimeofday( &tv, &tz );

  return ( tv.tv_sec * 1000 * 1000 ) + tv.tv_usec;
}

inline uint64 grab_tsc()
{
  register uint32 __a;
  register uint32 __d;

  __asm__ __volatile__( "rdtsc" : "=a"( __a ), "=d"( __d ) );

  return ( ( uint64 )( __d ) << 32 ) | 
         ( ( uint64 )( __a ) <<  0 ) ;
}

void dump( char* name, int n, int m, uint64 tod, uint64 tsc, int length, int trials )
{
  float t0 = ( ( float )( tod ) / ( float )( length ) ) / ( float )( trials );
  float t1 = ( ( float )( tsc ) / ( float )( length ) ) / ( float )( trials );

  printf( "%s [block=%d-bit,hash=%d-bit] : microsec/byte = %9.4f\n", name, n, m, t0 );
  printf( "%s [block=%d-bit,hash=%d-bit] :    cycle/byte = %9.4f\n", name, n, m, t1 );
}

void benchmark( int length, int trials )
{
  byte*  source  = NULL;
  byte*  target  = NULL;

  uint64 tod_beg = 0;
  uint64 tod_end = 0;
  uint64 tod_tot = 0;

  uint64 tsc_beg = 0;
  uint64 tsc_end = 0;
  uint64 tsc_tot = 0;

  lash_init();

  source = ( byte* )( malloc( length            ) );
  target = ( byte* )( malloc( lash_sizeof_dig() ) );

  for( int i = 0; i < trials; i++ )
  {
    for( int j = 0; j < length; j++ )
      source[ j ] = rand();

    tod_beg = grab_tod(); 
    tsc_beg = grab_tsc(); 
    lash_proc( target, source, length );
    tsc_end = grab_tsc(); 
    tod_end = grab_tod(); 

    tsc_tot += ( tsc_end - tsc_beg );
    tod_tot += ( tod_end - tod_beg );
  }

  free( source );
  free( target );

  lash_fini();

  int n = lash_sizeof_blk()*8;
  int m = lash_sizeof_dig()*8;

  dump( "LASH", n, m, tod_tot, tsc_tot, length, trials );
}

int main( int argc, char* argv[] ) 
{
  if( argc != 3 )
  {
    printf( "usage : benchmark <length> <trials>\n" );
    exit( 0 );
  }

  int length = atoi( argv[ 1 ] );
  int trials = atoi( argv[ 2 ] );

  benchmark( length, trials );
}
