#include "lash.h"

#define LASH_CONFIG_PRNG_INI (      54321UL )
#define LASH_CONFIG_PRNG_MOD ( 2147483647UL )

#if   defined( LASH_CONFIG_SIZE_160 )
#define LASH_CONFIG_SIZE_PART0
#elif defined( LASH_CONFIG_SIZE_256 )
#define LASH_CONFIG_SIZE_PART0
#define LASH_CONFIG_SIZE_PART1
#elif defined( LASH_CONFIG_SIZE_384 )
#define LASH_CONFIG_SIZE_PART0
#define LASH_CONFIG_SIZE_PART1
#define LASH_CONFIG_SIZE_PART2
#elif defined( LASH_CONFIG_SIZE_512 )
#define LASH_CONFIG_SIZE_PART0
#define LASH_CONFIG_SIZE_PART1
#define LASH_CONFIG_SIZE_PART2
#define LASH_CONFIG_SIZE_PART3
#endif

static int    m; //         output size
static int    n; //         input  size

static int    M; // aligned output size
static int    N; // aligned input  size

static byte*  H; // matrix

static inline uint32 prng( uint32 x )
{
  uint32 p[] = { 0x7FFFFFFFUL,
                 0x00000000UL };

  uint64 t   = ( uint64 )( x ) *
               ( uint64 )( x ) ;

  uint32 l   = ( uint32 )( t       ) & 0x7FFFFFFFUL;
  uint32 h   = ( uint32 )( t >> 31 ) & 0xFFFFFFFFUL;

  x  = l + h + 2 ;

  return x - p[ x < 0x7FFFFFFFUL ];
}

static inline uint32 padd32( uint32 x, uint32 y )
{
  uint32 t0 = x;
  uint32 t1 = y;

  uint32 t2 = t0 & 0x7F7F7F7FUL;
  uint32 t3 = t1 & 0x7F7F7F7FUL;
         t2 = t2 + t3;
         t3 = t0 ^ t1;
         t3 = t3 & 0x80808080UL;
         t3 = t3 ^ t2;

  return t3;
}

static inline uint64 padd64( uint64 x, uint64 y )
{
  uint64 t0 = x;
  uint64 t1 = y;

  uint64 t2 = t0 & 0x7F7F7F7F7F7F7F7FULL;
  uint64 t3 = t1 & 0x7F7F7F7F7F7F7F7FULL;
         t2 = t2 + t3;
         t3 = t0 ^ t1;
         t3 = t3 & 0x8080808080808080ULL;
         t3 = t3 ^ t2;

  return t3;
}

#if   defined( LASH_CONFIG_COMP_BYTE ) && defined( LASH_CONFIG_MATRIX_ALL )
#include "lash_x86_byte+all.h"
#elif defined( LASH_CONFIG_COMP_BYTE ) && defined( LASH_CONFIG_MATRIX_ROW )
#include "lash_x86_byte+row.h"
#elif defined( LASH_CONFIG_COMP_BYTE ) && defined( LASH_CONFIG_MATRIX_COL )
#include "lash_x86_byte+col.h"
#elif defined( LASH_CONFIG_COMP_WORD ) && defined( LASH_CONFIG_MATRIX_ALL )
#include "lash_x86_word+all.h"
#elif defined( LASH_CONFIG_COMP_WORD ) && defined( LASH_CONFIG_MATRIX_ROW )
#include "lash_x86_word+row.h"
#elif defined( LASH_CONFIG_COMP_WORD ) && defined( LASH_CONFIG_MATRIX_COL )
#include "lash_x86_word+col.h"
#elif defined( LASH_CONFIG_COMP_SIMD ) && defined( LASH_CONFIG_MATRIX_ALL )
#include "lash_x86_simd+all.h"
#elif defined( LASH_CONFIG_COMP_SIMD ) && defined( LASH_CONFIG_MATRIX_ROW )
#include "lash_x86_simd+row.h"
#elif defined( LASH_CONFIG_COMP_SIMD ) && defined( LASH_CONFIG_MATRIX_COL )
#include "lash_x86_simd+col.h"
#else
static inline void comp( byte* r, byte* s )
{

}
#endif
 
int lash_sizeof_dig()
{
  return ( m >> 1 );
}

int lash_sizeof_blk()
{
  return ( n >> 4 );
}

void lash_init()
{
  #if   defined( LASH_CONFIG_SIZE_160 )
  m =   40;
  n =  640;
  #elif defined( LASH_CONFIG_SIZE_256 )
  m =   64;
  n = 1024;
  #elif defined( LASH_CONFIG_SIZE_384 )
  m =   96;
  n = 1536;
  #elif defined( LASH_CONFIG_SIZE_512 )
  m =  128;
  n = 2048;
  #endif

  M = m;
  N = n;

  #if defined( LASH_CONFIG_COMP_SIMD )
  if( ( M % 16 ) != 0 )
    M = ( 16 - ( M % 16 ) ) + M;
  #endif

  #if   defined( LASH_CONFIG_MATRIX_ALL )
  H = ( byte* )( memalign( 16, M * N ) );

  uint32 g = LASH_CONFIG_PRNG_INI;

  for( int i = 0; i < n; i++ )
  {
    for( int j = 0; j < m; j++ )
    {
      H[ ( ( ( j - i + n ) % n ) * M ) + j ] = g;
    }

    g = prng( g );
  }
  #elif defined( LASH_CONFIG_MATRIX_ROW )
  H = ( byte* )( memalign( 16, M + N ) );

  uint32 g = LASH_CONFIG_PRNG_INI;

  for( int i = 0; i < n; i++ )
  {
    H[ i + 0 ] = g;

    g = prng( g );
  }

         g = LASH_CONFIG_PRNG_INI;

  for( int i = 0; i < m; i++ )
  {
    H[ i + n ] = g;

    g = prng( g );
  }
  #endif
}
 
void lash_fini()
{
  free( H );
}

void lash_proc( byte* d, byte* v, uint64 l )
{
  byte r[ M ];
  byte t[ M ];

  int p = l;

  memset( r, 0, M );
  memset( t, 0, M );

  while( p >= m )
  {
    comp( r, v );

    p -= m;
    v += m;
  }

  if( p )
    memcpy( t, v, p );

  t[ p ] = 0x80;

  comp( r, t );

  memset( t, 0, M );

  l = l << 3;

  t[ 0 ] = ( l       ) & 0xFF;
  t[ 1 ] = ( l >>  8 ) & 0xFF;
  t[ 2 ] = ( l >> 16 ) & 0xFF;
  t[ 3 ] = ( l >> 24 ) & 0xFF;
  t[ 4 ] = ( l >> 32 ) & 0xFF;
  t[ 5 ] = ( l >> 40 ) & 0xFF;
  t[ 6 ] = ( l >> 48 ) & 0xFF;
  t[ 7 ] = ( l >> 56 ) & 0xFF;

  comp( r, t );

  for( int i = 0; i < ( m >> 1 ); i++ )
  {
    int j = i << 1;

    t[ i ] = ( ( r[ j   ] & 0xF0 )      ) |
             ( ( r[ j+1 ] & 0xF0 ) >> 4 ) ;
  }

  memcpy( d, t, m >> 1 );
}
